# diarize_whisper/transcript.py

def combine_transcript_diarization(diarization, transcript, default_time: float = 0.0):
    """
    Combine les informations de transcription et de diarisation.
    """
    merged_segments = []
    diarization_segments = list(diarization.itertracks(yield_label=True))
    chunks = transcript.get("chunks", [])
    speakers = [None] * len(chunks)

    for i, chunk in enumerate(chunks):
        text_start, text_end = chunk.get("timestamp", (None, None))
        if text_start is not None and text_end is not None:
            for turn, _, speaker in diarization_segments:
                if turn.start <= text_start and turn.end >= text_end:
                    speakers[i] = speaker
                    break

    for i, chunk in enumerate(chunks):
        text_start, text_end = chunk.get("timestamp", (None, None))
        if text_start is None:
            if i > 0 and chunks[i-1].get("timestamp", (None, None))[1] is not None:
                text_start = chunks[i-1]["timestamp"][1]
            else:
                for j in range(i+1, len(chunks)):
                    ns, _ = chunks[j].get("timestamp", (None, None))
                    if ns is not None:
                        text_start = ns
                        break
                if text_start is None:
                    text_start = default_time
        if text_end is None:
            if i < len(chunks)-1 and chunks[i+1].get("timestamp", (None, None))[0] is not None:
                text_end = chunks[i+1]["timestamp"][0]
            else:
                text_end = text_start

        assigned_speaker = speakers[i]
        if assigned_speaker is None:
            if i > 0 and speakers[i-1] is not None:
                assigned_speaker = speakers[i-1]
            else:
                for j in range(i+1, len(chunks)):
                    if speakers[j] is not None:
                        assigned_speaker = speakers[j]
                        break
        if assigned_speaker is None:
            assigned_speaker = "UNKNOWN"

        merged_segments.append({
            "speaker": assigned_speaker,
            "start": round(text_start, 2),
            "end": round(text_end, 2),
            "text": chunk["text"]
        })

    return merged_segments

def refine_segments_with_context(segments, tolerance: float = 0.2):
    """
    Ajuste les segments en fonction du contexte (segments précédent et suivant).
    """
    refined = segments.copy()
    for i, seg in enumerate(refined):
        if i > 0:
            prev_seg = refined[i-1]
            if seg["start"] - prev_seg["end"] <= tolerance:
                if seg["text"].strip().startswith("-") or seg["text"].strip()[0] in {",", ".", ";", ":"}:
                    seg["speaker"] = prev_seg["speaker"]
        if i > 0 and i < len(refined)-1:
            prev_seg = refined[i-1]
            next_seg = refined[i+1]
            if len(seg["text"].split()) < 3 and prev_seg["speaker"] == next_seg["speaker"]:
                seg["speaker"] = prev_seg["speaker"]
    return refined

def reconstruct_dialogues(segments):
    """
    Reconstruit des dialogues en regroupant les segments continus du même locuteur.
    """
    dialogues = []
    if not segments:
        return dialogues

    current_speaker = segments[0]['speaker']
    current_start = segments[0]['start']
    current_end = segments[0]['end']
    current_text = [segments[0]['text']]

    for seg in segments[1:]:
        speaker = seg['speaker']
        if speaker != current_speaker:
            dialogues.append({
                'speaker': current_speaker,
                'start': current_start,
                'end': current_end,
                'text': " ".join(current_text)
            })
            current_speaker = speaker
            current_start = seg['start']
            current_text = [seg['text']]
        else:
            current_text.append(seg['text'])
        current_end = seg['end']

    dialogues.append({
        'speaker': current_speaker,
        'start': current_start,
        'end': current_end,
        'text': " ".join(current_text)
    })
    return dialogues



# diarize_whisper/__init__.py
from .audio import convert_to_wav_mono, split_wav_file
from .asr import get_device, get_torch_dtype, load_asr_pipeline
from .diarization import load_diarization_pipeline
from .transcript import (
    combine_transcript_diarization,
    refine_segments_with_context,
    reconstruct_dialogues,
)



# diarize_whisper/asr.py
import torch
from transformers import AutoModelForSpeechSeq2Seq, AutoProcessor, pipeline

def get_device() -> str:
    """
    Retourne le device ('cuda:0' ou 'cpu') selon la disponibilité.
    """
    return "cuda:0" if torch.cuda.is_available() else "cpu"

def get_torch_dtype() -> torch.dtype:
    """
    Retourne le type de tenseur adapté (float16 si CUDA est disponible).
    """
    return torch.float16 if torch.cuda.is_available() else torch.float32

def load_asr_pipeline(
    model_id: str,
    token: str,
    device: str,
    torch_dtype: torch.dtype,
    batch_size: int = 16,
    chunk_length_s: int = 25
):
    """
    Charge et retourne une pipeline ASR configurée.
    
    Args:
        model_id: Identifiant du modèle ASR.
        token: Token d'authentification HuggingFace.
        device: Device à utiliser.
        torch_dtype: Type de tenseur.
        batch_size: Taille de batch.
        chunk_length_s: Durée des segments en secondes.
    
    Returns:
        Pipeline de transcription automatique.
    """
    model = AutoModelForSpeechSeq2Seq.from_pretrained(
        model_id,
        torch_dtype=torch_dtype,
        low_cpu_mem_usage=True,
        use_safetensors=True,
        token=token
    )
    model.to(device)

    processor = AutoProcessor.from_pretrained(model_id, token=token)
    asr_pipeline_instance = pipeline(
        "automatic-speech-recognition",
        model=model,
        tokenizer=processor.tokenizer,
        feature_extractor=processor.feature_extractor,
        batch_size=batch_size,
        chunk_length_s=chunk_length_s,
        torch_dtype=torch_dtype,
        device=device,
    )
    return asr_pipeline_instance



# diarize_whisper/cli.py
import os
import argparse
from diarize_whisper.audio import convert_to_wav_mono, split_wav_file
from diarize_whisper.asr import get_device, get_torch_dtype, load_asr_pipeline
from diarize_whisper.diarization import load_diarization_pipeline
from diarize_whisper.transcript import (
    combine_transcript_diarization,
    refine_segments_with_context,
    reconstruct_dialogues,
)

def main():
    parser = argparse.ArgumentParser(description="Pipeline ASR et Diarisation")
    parser.add_argument("--input", type=str, required=True, help="Chemin vers le fichier audio d'entrée")
    parser.add_argument("--output", type=str, default="resultats_dialogues.txt", help="Fichier de sortie pour les dialogues")
    parser.add_argument("--chunk_duration", type=int, default=25, help="Durée des chunks en secondes")
    parser.add_argument("--asr_model", type=str, default="openai/whisper-large-v3-turbo", help="Modèle ASR à utiliser")
    parser.add_argument("--diarization_model", type=str, default="pyannote/speaker-diarization-3.1", help="Modèle de diarisation à utiliser")
    parser.add_argument("--token", type=str, required=True, help="Token d'authentification HuggingFace")
    parser.add_argument('--num_speakers', type=int, required=False, help="Nombre de speakers")
    parser.add_argument('--min_speakers', type=int, help="Nombre de speakers minumum")
    parser.add_argument("--max_speakers", type=int, help="Nombre de speakers maximum")
    args = parser.parse_args()

    device = get_device()
    torch_dtype = get_torch_dtype()

    asr_pipeline = load_asr_pipeline(args.asr_model, args.token, device, torch_dtype, chunk_length_s=args.chunk_duration)
    diarization_pipeline = load_diarization_pipeline(args.diarization_model, device="cuda")

    # Conversion et découpage du fichier audio
    wav_file = convert_to_wav_mono(args.input, "sample.wav")
    split_dir = "output_chunks"
    os.makedirs(split_dir, exist_ok=True)
    split_wav_file(wav_file, os.path.join(split_dir, "audio"), segment_length_ms=args.chunk_duration * 1000)

    # Diarisation globale sur le fichier complet
    dia_result = diarization_pipeline(wav_file, num_speakers=args.num_speakers, min_speakers=args.min_speakers, max_speakers=args.max_speakers)

    all_merged_segments = []
    files = sorted(os.listdir(split_dir))  # Assurer l'ordre correct

    for idx, file in enumerate(files):
        file_path = os.path.join(split_dir, file)
        try:
            transcript_result = asr_pipeline(file_path, return_timestamps="word")
        except TypeError as e:
            if "'<=' not supported" in str(e):
                print(f"Erreur détectée pour {file}: {e}. Relance de la pipeline avec return_timestamps=True")
                transcript_result = asr_pipeline(file_path, return_timestamps=True)
            else:
                raise e

        # Ajustement des timestamps selon l'offset du chunk
        offset = idx * args.chunk_duration
        for chunk in transcript_result.get("chunks", []):
            start, end = chunk["timestamp"]
            if start is not None and end is not None:
                chunk["timestamp"] = (start + offset, end + offset)

        merged_segments = combine_transcript_diarization(dia_result, transcript_result)
        all_merged_segments.extend(merged_segments)
        all_merged_segments = refine_segments_with_context(all_merged_segments, tolerance=0.2)

    dialogues = reconstruct_dialogues(all_merged_segments)
    with open(args.output, "w", encoding="utf-8") as f:
        for dialogue in dialogues:
            line = f"[{dialogue['speaker']} ({dialogue['start']} - {dialogue['end']})] {dialogue['text']}\n"
            print(line)
            f.write(line)

if __name__ == "__main__":
    main()



# diarize_whisper/diarization.py
import torch
from pyannote.audio import Pipeline

def load_diarization_pipeline(model_id: str, device: str = "cuda"):
    """
    Charge et retourne une pipeline de diarisation.
    
    Args:
        model_id: Identifiant du modèle de diarisation.
        device: Device à utiliser (par défaut "cuda").
    
    Returns:
        Pipeline de diarisation.
    """
    diarization = Pipeline.from_pretrained(model_id)
    return diarization.to(torch.device(device))



# setup.py
from setuptools import setup, find_packages
import os

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()
    


setup(
    name="diarize_whisper",
    version="0.1.2",
    packages=find_packages(),
    install_requires=[
        "pydub",
        "torch",
        "pyannote.audio",
        "transformers",
    ],
    author="Samy Le Galloudec",
    author_email="samy.legalloudec@gmail.com",
    description="Librairie pour la transcription ASR et la diarisation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    # url="https://github.com/votrecompte/diarize_whisper",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
    entry_points={
        'console_scripts': [
            'diarize_whisper=diarize_whisper.cli:main',
        ],
    },
)



# diarize_whisper/audio.py
from pydub import AudioSegment
import math

def convert_to_wav_mono(input_file: str, output_file: str = "temp_audio.wav") -> str:
    """
    Convertit un fichier audio en WAV mono 16kHz.
    """
    audio = AudioSegment.from_file(input_file)
    audio = audio.set_channels(1).set_frame_rate(16000)
    audio.export(output_file, format="wav")
    return output_file

def split_wav_file(input_file: str, output_prefix: str, segment_length_ms: int = 30*1000) -> None:
    """
    Découpe un fichier WAV en segments d'une durée donnée.
    """
    audio = AudioSegment.from_wav(input_file)
    total_length = len(audio)
    num_segments = math.ceil(total_length / segment_length_ms)

    for i in range(num_segments):
        start = i * segment_length_ms
        end = min((i + 1) * segment_length_ms, total_length)
        segment = audio[start:end]
        output_file = f"{output_prefix}_part_{i+1}.wav"
        segment.export(output_file, format="wav")
        print(f"Segment {i+1} exporté vers {output_file}")
